module Enviz_SketchUp_Plugin
    module Enviz_Sketchup_Helpers

        @last_render_settings = {}

        def self.convert_to_inches(value)
            model = Sketchup.active_model
            units = model.options["UnitsOptions"]["LengthUnit"]
            puts "UNITS = #{units}" if DEBUG_MODE
            case units
            when 0 # Model units are inches
            value / 0.0254
            when 1 # Model units are feet
            value / 0.3048
            when 2 # Model units are millimeters
            value / 0.001
            when 3 # Model units are centimeters
            value / 0.01
            when 4 # Model units are meters
            value / 1
            else
            value / 1# default to meters if unknown
            end
        end

        def self.convert_to_meters(value)
            model = Sketchup.active_model
            units = model.options["UnitsOptions"]["LengthUnit"]
            puts "UNITS = #{units}" if DEBUG_MODE
            case units
            when 0 # Model units are inches
            value * 0.0254
            when 1 # Model units are feet
            value * 0.3048
            when 2 # Model units are millimeters
            value * 0.001
            when 3 # Model units are centimeters
            value * 0.01
            when 4 # Model units are meters
            value * 1
            else
            value * 1# default to meters if unknown
            end
        end

        def self.raycast_face(x, y, view)
            ph = view.pick_helper
            ph.do_pick(x, y)
            best = ph.picked_face  # This is often used to get the most front-most entity
        
            if best #.is_a?(Sketchup::Face)
                return best
            else
                # puts "No face selected"
                return nil
            end
        end    

        def self.raycast_hit_point(x, y, view)
            input_point = Sketchup::InputPoint.new
            input_point.pick(view, x, y)
            
            # The input_point will now contain the correct position if it hit something
            if input_point.valid? && input_point.face
                hit_point = input_point.position
                puts "Hit point: #{hit_point.to_s}" if DEBUG_MODE
                return hit_point
            else
                puts "No valid intersection found or no face under cursor" if DEBUG_MODE
                UI.messagebox("Failed to select the Entity. Please try again")
                return nil
            end
        end

        def self.set_backfaces_red
            ro = Sketchup.active_model.rendering_options
            @last_render_settings['FaceBackColor'] = ro['FaceBackColor']
            @last_render_settings['RenderMode'] = ro['RenderMode']

            puts "Setting backfaces to red" if DEBUG_MODE
            ro['FaceBackColor'] = 'red'
            ro['RenderMode'] = 5
        end

        def self.reset_render_options
            ro = Sketchup.active_model.rendering_options
            @last_render_settings.each do |key, value|
                ro[key] = value
            end
        end

        def face_normal_direction(face)
           
            # Get the active view and its camera
            view = Sketchup.active_model.active_view
            camera = view.camera
            
            # Get the normal vector of the face
            normal = face.normal
            
            # Get the direction the camera is pointing
            camera_direction = camera.direction
            
            # Calculate the dot product of the normal and the camera direction
            dot_product = normal.dot(camera_direction)
            
            # Check the dot product to determine the direction relative to the camera
            if dot_product > 0
                return false
            else
                return true
            end
        
        end

    end
end
