Sketchup::require('enviz_sketchup_plugin/enviz_sketchup_helpers')

# Face flipper tool for correcting flipped normals
# When activated it will change the view mode to monochrome so that backfaces are visible
# Then the user can single left-click on faces to flip them
module Enviz_SketchUp_Plugin
  module Enviz_Small_Tools
      
      include Enviz_Sketchup_Helpers   

      class Tool_Flip_All

          def self.flip_all_normals

              Enviz_Sketchup_Helpers.set_backfaces_red

              # Get the active SketchUp model
              model = Sketchup.active_model
              
              # Start an operation which can be undone in one step
              model.start_operation("Flip All Normals", true)
              
              # Access the entities in the model's active entities context
              entities = model.active_entities
              
              # Recursively flip normals of all faces in the entity collection
              flip_normals_in_collection(entities)
              
              # Commit the operation
              model.commit_operation
          end
          
          def self.flip_normals_in_collection(entities, on_complete: nil)
              model = Sketchup.active_model
              queue = entities.to_a.dup # Initialize queue with a shallow copy of entities
              processed = {} # To keep track of processed entities to avoid reprocessing
            
              process_entities = lambda do
                start_time = Time.now
            
                while queue.any? && Time.now - start_time < 0.25 # Allow a longer processing window
                  entity = queue.shift
                  next if processed[entity.object_id] # Skip if already processed
            
                  processed[entity.object_id] = true
            
                  if entity.is_a?(Sketchup::Face)
                    entity.reverse!
                  elsif entity.is_a?(Sketchup::Group) || entity.is_a?(Sketchup::ComponentInstance)
                    entity.definition.entities.each { |e| queue.push(e) unless processed[e.object_id] }
                  end
                end
            
                if queue.any?
                  # Schedule less frequent but longer processing chunks
                  UI.start_timer(0.1, false) { process_entities.call }
                else
                  on_complete.call if on_complete.is_a?(Proc)
                end
              end
            
              process_entities.call
            end

      end
  end
end