require 'sketchup'
Sketchup::require('enviz_sketchup_plugin/enviz_sketchup_helpers')
Sketchup::require('enviz_sketchup_plugin/enviz_small_tools')
Sketchup::require('enviz_sketchup_plugin/enviz_ui')
Sketchup::require('enviz_sketchup_plugin/teleport_handler')
Sketchup::require('enviz_sketchup_plugin/tool_flipper')
Sketchup::require('enviz_sketchup_plugin/tool_validator')
Sketchup::require('enviz_sketchup_plugin/tool_dollhouse_creator')
Sketchup::require('enviz_sketchup_plugin/tool_exporter')

module Enviz_SketchUp_Plugin
    
    module Enviz_UI
        include Enviz_Sketchup_Helpers
        include Enviz_Small_Tools
        include Teleports
        include Dollhouse
        include Validator
        

        @current_state = "main"
        @dialog = nil

        # STATIC #

        @dialog_position = { left: 1, top: 82 }  # Default position

        # Method to create a new dialog
        def self.create_dialog(winWidth, winHeight)
            close_dialog if @dialog

            #can_scroll = false

            puts "WIN HEIGHT - #{winHeight}" if DEBUG_MODE
            #if winHeight == 500
                #can_scroll = true
                #puts "SCROLLABLE" if DEBUG_MODE
            #end

            @dialog = UI::HtmlDialog.new(
                dialog_title: "Enviz Sketchup Plugin",
                scrollable: true,
                resizable: false,
                width: winWidth,
                height: winHeight,
                left: @dialog_position[:left],
                top: @dialog_position[:top],
            )

            create_callbacks unless @callbacks_created
            @callbacks_created = true
            @dialog
        end

        def self.teleport_order_updated()

            puts "TELEPORT ORDER UPDATED" if DEBUG_MODE

        end 

        def self.close_dialog
            return unless @dialog

            detach_selection_observer  
        
            @dialog.close
            @dialog = nil
        end

        def self.detach_selection_observer
            model = Sketchup.active_model
            selection = model.selection
            begin
                selection.remove_observer(@simple_observer) if @simple_observer
            rescue StandardError => e
                puts "Error detaching observer: #{e.message}" if DEBUG_MODE
            ensure
                @simple_observer = nil  # Ensure the reference is cleared to avoid memory leaks
            end
        end

        # Default values because
        @current_menu = "main"
        winWidth = 100
        winHeight = 100
        @validation_run = false

        def self.show_processing_overlay(show, delay)

            sleep(delay)

            @dialog.execute_script("toggleOverlay(#{show})") 

        end

        def self.attach_selection_observer
            model = Sketchup.active_model
            selection = model.selection
            @simple_observer = SimpleSelectionObserver.new
            selection.add_observer(@simple_observer)
        end


        @registered_callbacks = {}

        def self.create_callbacks
            callbacks = {
                "handle_rotation_change" => lambda { |action_context, rotation_value|
                    Teleports.handle_rotation_change(rotation_value)
                },
                "say" => lambda { |action_context, say_val|
                    puts "#{say_val}" if DEBUG_MODE
                },
                "reorder_teleports" => lambda { |action_context, item_to_move, new_position|
                    puts "Reordering teleports: #{item_to_move} <-> #{new_position}" if DEBUG_MODE
                    Teleports::TeleportHandler.reorder_teleports(item_to_move, new_position)
                    goto_teleport_manager(true)
                },
                "teleport_preview" => lambda { |action_context, order|
                    Teleports.preview(order)
                },
                "delete_teleport" => lambda { |action_context, order|
                    Teleports::TeleportHandler.delete_teleport(order)
                    goto_teleport_creator(true)
                },
                "teleport_preview_manager" => lambda { |action_context, order|
                    Teleports.preview(order)
                    goto_teleport_manager(true)
                },
                "delete_teleport_manager" => lambda { |action_context, order|
                    Teleports::TeleportHandler.delete_teleport(order)
                    goto_teleport_manager(true)
                },
                "update_teleport_orders" => lambda { |action_context, new_orders|
                    puts "I GUESS IT WORKED" if DEBUG_MODE
                    goto_teleport_manager(true)
                },
                "set_teleport_name" => lambda { |action_context, text|
                    Teleports.set_teleport_name(text)
                    goto_teleport_creator(true)
                },
                "summon_fix_faces" => lambda { |action_context|
                    goto_fix_faces(false)
                },
                "summon_dollhouse_creator" => lambda { |action_context|
                    goto_dollhouse_creator(false)
                },
                "summon_main" => lambda { |action_context|
                    goto_main(false)
                },
                "summon_teleport_creator" => lambda { |action_context|


                    goto_teleport_creator(false)
                },
                "tool_flip_all" => lambda { |action_context|
                    show_processing_overlay(true, 0)
                    Sketchup.active_model.select_tool(nil)
                    Enviz_Small_Tools::Tool_Flip_All.flip_all_normals()
                    goto_fix_faces(true)
                    show_processing_overlay(false, 0.5)
                },
                "tool_flipper" => lambda { |action_context|
                    show_processing_overlay(true, 0)
                    if is_tool_active(Tool_Flipper)
                        Sketchup.active_model.select_tool(nil)
                    else
                        Sketchup.active_model.select_tool(Tool_Flipper.new)
                    end
                    goto_fix_faces(true)
                    show_processing_overlay(false, 0.5)
                },
                "tool_correct_object" => lambda { |action_context|
                    show_processing_overlay(true, 0)
                    if is_tool_active(Tool_CorrectObject)
                        Sketchup.active_model.select_tool(nil)
                    else
                        Sketchup.active_model.select_tool(Tool_CorrectObject.new)
                    end
                    goto_fix_faces(true)
                    show_processing_overlay(false, 0.5)
                },
                "tool_teleport_creator" => lambda { |action_context|
                    if is_tool_active(Teleports::Tool_Teleport_Creator)
                        Sketchup.active_model.select_tool(nil)
                    else
                        Sketchup.active_model.select_tool(Teleports::Tool_Teleport_Creator.new)
                    end
                    goto_teleport_creator(true)
                },
                "tool_dollhouse_creator_pivot" => lambda { |action_context|
                    if is_tool_active(Tool_Dollhouse_Creator_Pivot)
                        Sketchup.active_model.select_tool(nil)
                    else
                        Sketchup.active_model.select_tool(Tool_Dollhouse_Creator_Pivot.new)
                    end
                    goto_dollhouse_creator(true)
                },
                "tool_dollhouse_creator_view" => lambda { |action_context|
                    if is_tool_active(Tool_Dollhouse_Creator_View)
                        Sketchup.active_model.select_tool(nil)
                    else
                        Sketchup.active_model.select_tool(Tool_Dollhouse_Creator_View.new)
                    end
                    goto_dollhouse_creator(true)
                },
                "tool_confirm_dollhouse" => lambda { |action_context|
                    Dollhouse.set_dollhouse_view()
                    Sketchup.active_model.select_tool(nil)
                    goto_main(false)
                },
                "update_dollhouse_view" => lambda { |action_context, rotation, distance, height|
                    Dollhouse.update_dollhouse_view(rotation, distance, height)
                },
                "validate_model" => lambda { |action_context|
                    show_processing_overlay(true, 0)
                    Validator.validate_model
                    @validation_run = true
                    goto_main(true)
                    show_processing_overlay(false, 0.5)
                },
                "export_model" => lambda { |action_context|
                    show_processing_overlay(true, 0)
                    Exporter.export_model
                    show_processing_overlay(false, 0.5)
                },
                "manage_teleports" => lambda { |action_context|
                    show_processing_overlay(true, 0)
                    goto_teleport_manager(false)
                    show_processing_overlay(false, 0.5)
                }
            }
        
            callbacks.each do |action, proc|
            # Register the callback only if it hasn't been registered before
            # unless @registered_callbacks[action]
                @dialog.add_action_callback(action, &proc)
                # @registered_callbacks[action] = true
            # end 
            end
            @dialog.set_on_closed { @callbacks_created = false}
        end

        # Store the dialog reference
        Sketchup.active_model.select_tool(nil)
        @dialog = create_dialog(winWidth, winHeight)

        def self.summon_dialog(content, winWidth, winHeight)

            Sketchup.active_model.select_tool(nil)

            if winHeight > 500
                winHeight = 500
            end
            
            @dialog.set_size(winWidth, winHeight)

            # Hook into the move event to save the dialog's current position
            # @dialog.add_action_callback("dialogMoved") do |action_context, left, top|
            #     puts "MOVED"
            #     @dialog_position = { left: left.to_i, top: top.to_i }
            # end

            # Set new HTML content
            @dialog.set_html(content)

            create_callbacks unless @callbacks_created
            @callbacks_created = true

            @dialog.show
        end

        def self.update_dialog(content, winWidth, winHeight)
            
            if winHeight > 500
                winHeight = 500
            end

            @dialog.set_size(winWidth, winHeight)

            # Hook into the move event to save the dialog's current position
            # @dialog.add_action_callback("dialogMoved") do |action_context, left, top|
            #     puts "MOVED"
            #     @dialog_position = { left: left.to_i, top: top.to_i }
            # end

            # Set new HTML content
            @dialog.set_html(content)

            create_callbacks unless @callbacks_created
            @callbacks_created = true

            @dialog.show
        end

        if defined?(ENVIZSTYLES)
            Object.send(:remove_const, :ENVIZSTYLES)
        end

        ENVIZSTYLES = <<-CSS
            
            .container {
                display: flex;
                align-items: center;
                justify-content: flex-start; /* Adjusted to space-between to allow the range slider to take up the remaining space on the right */
                gap: 10px;
                width: 100%; /* Ensure the container takes full width */
            }    

            .primary-button {
                /*typography*/
                color: #6332A4;
                text-align: center;
                font-family: "Helvetica";
                font-size: 14px;
                font-style: normal;
                font-weight: 700;
                line-height: 100%; /* 16px */
                letter-spacing: 0.32px;
                /*style*/
                display: flex;
                padding: 16px 16px;
                justify-content: center;
                align-items: center;
                gap: 8px;
                align-self: stretch;
                border: none;
                border-radius: 10px;
                border-radius: radius/m;
                background: var(--primary-button-fill-default, #F5EEFE);
                width: 100%; /* Stretch to full width */
                box-sizing: border-box;
                margin-bottom: 8px;
                margin-top: 8px;
                filter: blur(0.25px);
            }
            .primary-button:hover {
                /*typography*/
                background-color: #8673FC; color: #FFFFFF;
            }

            .trash-button {
                /*typography*/
                color: #E55555;
                text-align: center;
                font-family: "Helvetica";
                font-size: 20px;
                font-style: normal;
                font-weight: 700;
                line-height: 100%; /* 16px */
                letter-spacing: 0.32px;
                /*style*/
                display: flex;
                padding: 16px 16px;
                justify-content: center;
                align-items: center;
                gap: 8px;
                align-self: stretch;
                border: none;
                border-radius: 10px;
                border-radius: radius/m;
                background: var(--primary-button-fill-default, #FEE2E2);
                width: 86px; /* Stretch to full width */
                box-sizing: border-box;
                margin-bottom: 8px;
                margin-top: 8px;
                filter: blur(0.25px);
            }
            .trash-button:hover {
                /*typography*/
                background-color: #8673FC; color: #FFFFFF;
            }

            .dropdown {
                color: #000000;
                font-family: "Helvetica";
                font-size: 14;
                text-align: left;
                font-style: normal;
                font-weight: 400;
                line-height: 150%; /* 24px */
                filter: blur(0.25px);
                padding: 15px 15px;
                border: 1px solid #ccc;
                border-radius: 8px;
            }

            .text_input {
                color: #000000;
                font-family: "Helvetica";
                font-size: 14;
                text-align: left;
                font-style: normal;
                font-weight: 400;
                line-height: 150%; /* 24px */
                filter: blur(0.25px);
                padding: 16px 15px;
                border: 1px solid #ccc;
                border-radius: 8px;
            }

            .primary-button-tool_on {
                /*typography*/
                color: #F5EEFE;
                text-align: center;
                font-family: "Helvetica";
                font-size: 14px;
                font-style: normal;
                font-weight: 700;
                line-height: 100%; /* 16px */
                letter-spacing: 0.32px;
                /*style*/
                display: flex;
                padding: 16px 16px;
                justify-content: center;
                align-items: center;
                gap: 8px;
                align-self: stretch;
                border: none;
                border-radius: 10px;
                border-radius: radius/m;
                background: var(--primary-button-fill-default, #6332A4);
                width: 100%; /* Stretch to full width */
                box-sizing: border-box;
                margin-bottom: 8px;
                margin-top: 8px;
                filter: blur(0.25px);
            }
            .primary-button-tool_on:hover {
                /*typography*/
                background-color: #8673FC; color: #FFFFFF;
            }

            .primary-button:disabled {
                /*typography*/
                color: #FFFFFF;
                text-align: center;
                font-family: "Helvetica";
                font-size: 14px;
                font-style: normal;
                font-weight: 700;
                line-height: 100%; /* 16px */
                letter-spacing: 0.32px;
                /*style*/
                display: flex;
                padding: 16px 16px;
                justify-content: center;
                align-items: center;
                gap: 8px;
                align-self: stretch;
                border: none;
                border-radius: 10px;
                border-radius: radius/m;
                background: var(--primary-button-fill-default, #474747);
                width: 100%; /* Stretch to full width */
                box-sizing: border-box;
                margin-bottom: 8px;
                margin-top: 8px;
                filter: blur(0.25px);
            }

            .loading-button {
                /* Typography */
                color: #FFFFFF;
                text-align: center;
                font-family: "Helvetica";
                font-size: 14px;
                font-weight: 700;
                line-height: 100%; /* 16px, adjust if needed */
                letter-spacing: 0.32px;

                /* Style & Layout */
                display: flex;
                justify-content: center;
                align-items: center;
                padding: 16px 16px;
                border: none;
                border-radius: 10px;
                background: rgba(99, 50, 164, 0.5); /* Semi-transparent gray */
                width: 100%; /* Stretch to full width */
                height: 100%;
                box-sizing: border-box;
                margin-top: 8px;
                margin-bottom: 8px;
                gap: 8px; /* Spacing between inner elements if any */
                filter: blur(0.25px);
            }
            .loading-button:hover {
                /*typography*/
                background-color: #8673FC; color: #FFFFFF;
            }

            .primary-button:focus {
                /*typography*/
                color: #FFFFFF;
                text-align: center;
                font-family: "Helvetica";
                font-size: 14px;
                font-style: normal;
                font-weight: 700;
                line-height: 100%; /* 16px */
                letter-spacing: 0.32px;
                /*style*/
                display: flex;
                padding: 16px 16px;
                justify-content: center;
                align-items: center;
                gap: 8px;
                align-self: stretch;
                border: none;
                border-radius: 10px;
                border-radius: radius/m;
                background: var(--primary-button-fill-default, #201E4E);
                width: 100%; /* Stretch to full width */
                box-sizing: border-box;
                margin-bottom: 8px;
                margin-top: 8px;
                filter: blur(0.25px);
            }

            /*buttons*/
            .final-button {
                /*typography*/
                color: #FFFFFF;
                text-align: center;
                font-family: "Helvetica";
                font-size: 14px;
                font-style: normal;
                font-weight: 700;
                line-height: 100%; /* 16px */
                letter-spacing: 0.32px;
                /*style*/
                display: flex;
                padding: 16px 16px;
                justify-content: center;
                align-items: center;
                gap: 8px;
                align-self: stretch;
                border: none;
                border-radius: 10px;
                border-radius: radius/m;
                background: var(--primary-button-fill-default, #201E4E);
                width: 100%; /* Stretch to full width */
                box-sizing: border-box;
                margin-bottom: 8px;
                margin-top: 8px;
                filter: blur(0.25px);
            }
            .loading-button:hover {
                /*typography*/
                background-color: #8673FC; color: #FFFFFF;
            }

            .final-button:focus {
                /*typography*/
                color: #6332A4;
                text-align: center;
                font-family: "Helvetica";
                font-size: 14px;
                font-style: normal;
                font-weight: 700;
                line-height: 100%; /* 16px */
                letter-spacing: 0.32px;
                /*style*/
                display: flex;
                padding: 16px 16px;
                justify-content: center;
                align-items: center;
                gap: 8px;
                align-self: stretch;
                border: none;
                border-radius: 10px;
                border-radius: radius/m;
                background: var(--primary-button-fill-default, #F5EEFE);
                width: 100%; /* Stretch to full width */
                box-sizing: border-box;
                margin-bottom: 8px;
                margin-top: 8px;
                filter: blur(0.25px);
            }

            .back-button {
                /*typography*/
                color: #6332A4;
                text-align: center;
                font-family: "Helvetica";
                font-size: 14px;
                font-style: normal;
                font-weight: 700;
                line-height: 100%; /* 16px */
                letter-spacing: 0.32px;
                /*style*/
                display: flex;
                padding: 6px 16px;
                justify-content: left;
                align-items: left;
                gap: 8px;
                align-self: stretch;
                border: none;
                border-radius: 10px;
                border-radius: radius/m;
                background: var(--primary-button-fill-default, #F5EEFE);
                width: auto; /* Stretch to full width */
                box-sizing: border-box;
                margin-bottom: 8px;
                margin-top: 8px;
                filter: blur(0.25px);
            }
            .back-button:hover {
                /*typography*/
                background-color: #8673FC; color: #FFFFFF;
            }

            .back-button:focus {
                /*typography*/
                color: #FFFFFF;
                text-align: center;
                font-family: "Helvetica";
                font-size: 14px;
                font-style: normal;
                font-weight: 700;
                line-height: 100%; /* 16px */
                letter-spacing: 0.32px;
                /*style*/
                display: flex;
                padding: 6px 16px;
                justify-content: left;
                align-items: left;
                gap: 8px;
                align-self: stretch;
                border: none;
                border-radius: 10px;
                border-radius: radius/m;
                background: var(--primary-button-fill-default, #201E4E);
                width: auto; /* Stretch to full width */
                box-sizing: border-box;
                margin-bottom: 8px;
                margin-top: 8px;
                filter: blur(0.25px);
            }
            
            .main-heading {
                /* Headings/H5 */
                font-family: "Helvetica";
                font-size: 24px;
                font-style: bold;
                font-weight: 700;
                line-height: 150%; /* 33.6px */
                margin-bottom: 4px;
                filter: blur(0.25px);
            }

            .sub-heading {
                color: #000000;
                font-family: "Helvetica";
                font-size: 18;
                font-style: normal;
                font-weight: 700;
                line-height: 150%; /* 28px */
                filter: blur(0.25px);
            }

            .main-text {
                color: #000000;
                font-family: "Helvetica";
                font-size: 14;
                font-style: normal;
                font-weight: 400;
                line-height: 150%; /* 24px */
                filter: blur(0.25px);
            }

            .teleport-name {
                flex: 1;
                text-align: left;
                color: #000000;
                font-family: "Helvetica";
                font-size: 18;
                font-style: normal;
                font-weight: 700;
                line-height: 150%; /* 24px */
                filter: blur(0.25px);
                display: inline-block;
                max-width: 150px; /* Adjust the width as needed */
                white-space: nowrap;
                overflow: hidden;
                text-overflow: ellipsis;
            }

            .side-text {
                color: #000000;
                font-family: "Helvetica";
                font-size: 14;
                font-style: normal;
                font-weight: 400;
                line-height: 150%; /* 24px */
                filter: blur(0.25px);
                white-space: nowrap; /* Prevents the text from wrapping */
                flex-shrink: 0; /* Prevents the label from shrinking */
                padding-right: 10px; /* Adds some spacing between label and slider */
            }

            .main-text br {
                line-height: 0.4;     /* Reduces the line height to 0 */
            }

            .squishy-text {
                color: #000000;
                text-align: "left";
                font-family: "Helvetica";
                font-size: 12;
                font-style: normal;
                font-weight: 400;
                line-height: 150%; /* 24px */
                filter: blur(0.25px);
            }

            .colored-line {
                border: none;          /* Removes the default border */
                height: 1px;           /* Sets the thickness of the line */
                background-color: #E4D4F8; /* Sets the color of the line */
                margin-bottom: 16px;
                margin-top: 16px;
            }

            .make-border {

                margin-left: 16px;
                margin-right: 16px;

            }

            .slider-container {
                display: flex;
                flex-direction: column;
                align-items: center;
                margin-bottom: 20px; /* Spacing between each slider container */
            }

            /* Styling the slider input */
            input[type=range] {
                -webkit-appearance: none; /* Override default CSS styles */
                appearance: none;
                width: 50%; /* Full-width */
                text-align: right;
                align: right;
                height: 2px; /* Specified height */
                background: #ddd; /* Grey background */
                outline: none; /* Remove outline */
                opacity: 0.7; /* Set transparency (cross-browser) */
                -webkit-transition: .2s; /* 0.2 seconds transition on hover */
                transition: opacity .2s;
            }

            /* Mouse-over effects */
            input[type=range]:hover {
                opacity: 1; /* Fully opaque on hover */
            }

            /* The slider handle (the thumb) */
            input[type=range]::-webkit-slider-thumb {
                -webkit-appearance: none; /* Override default look */
                appearance: none;
                width: 20px; /* Width of the handle */
                height: 20px; /* Height of the handle */
                background: #6332A4; /* Purple background */
                cursor: pointer; /* Cursor on hover */
                border-radius: 50%; /* Circular shape */
            }

            input[type=range]::-moz-range-thumb {
                width: 20px; /* Width of the handle */
                height: 20px; /* Height of the handle */
                background: #6332A4; /* Purple background */
                cursor: pointer; /* Cursor on hover */
                border-radius: 50%; /* Circular shape */
            }

            /* Make the thumb larger and colored on focus */
            input[type=range]:focus::-webkit-slider-thumb {
                background: #b300b3; /* Darker purple background */
            }

            input[type=range]:focus::-moz-range-thumb {
                background: #b300b3; /* Darker purple background */
            }

            .range_slider_container {
                flex-grow: 1; /* Allows the container to grow and fill available space */
                display: flex;
                align-items: center;
            }

            .range_slider {
                -webkit-appearance: none;
                appearance: none;
                width: 70%; /* Slider takes up 70% of the .range_slider_container */
                max-width: 400px; /* Adjust this value based on your layout's max size */
                background: #ddd;
                outline: none;
                transition: opacity .2s;
            }

            .range_slider:hover {
                opacity: 1; /* Fully opaque on hover */
            }

            .range_slider::-webkit-slider-thumb {
                -webkit-appearance: none; /* Override default look */
                appearance: none;
                width: 20px; /* Width of the handle */
                height: 20px; /* Height of the handle */
                background: #6332A4; /* Purple background */
                cursor: pointer; /* Cursor on hover */
                border-radius: 50%; /* Circular shape */
            }

            .range_slider::-moz-range-thumb {
                width: 20px; /* Width of the handle */
                height: 20px; /* Height of the handle */
                background: #6332A4; /* Purple background */
                cursor: pointer; /* Cursor on hover */
                border-radius: 50%; /* Circular shape */
            }

            #overlay {
                position: fixed; /* Positioning relative to the viewport */
                top: 0;
                left: 0;
                width: 100%; /* Full width */
                height: 100%; /* Full height */
                background-color: rgba(1, 1, 1, 0.5); 
                z-index: 1000; /* High z-index to be on top of other elements */
                display: None;
                justify-content: center !important;
                align-items: center !important;
                pointer-events: all;
            }

            .disable-mouse {
                pointer-events: none;
            }

        CSS

        # ------------- SECTIONS ------------- #

        # ------------- HELPERS -------------- #

        def self.is_tool_active(test_tool)

            model = Sketchup.active_model

            current_tool = model.tools.active_tool

            if current_tool.nil?
                puts "nil" if DEBUG_MODE
                return false
            elsif current_tool.is_a?(test_tool)
                puts "is active tool" if DEBUG_MODE
                return true
            else
                puts "inactive" if DEBUG_MODE
                return false
            end

        end

        # ------------- HELPERS -------------- #

        # ------------------------------------------- MAIN --------------------------------------------- #
        def self.goto_main(is_update)

            # ✘
            # ✔
            # ☑
            # ☒

            puts Validator.poly_string if DEBUG_MODE

            # if Validator.is_valid
                # Validator.validate_model
            # end

            if !is_update
                @validation_run = false
            end
            
            @current_menu = "main"

            winWidth = 450;
            winHeight = 860;

            teleport_text = "Place teleport points"
            dollhouse_text = "Setup dollhouse view"

            teleport_tick = "<span style='color: red;'>✘</span>"
            dollhouse_tick = "<span style='color: red;'>✘</span>"
            poly_tick = "<span style='color: red;'>✘</span>"
            surface_tick = "<span style='color: red;'>✘</span>"
            object_tick = "<span style='color: red;'>✘</span>"
            texture_tick = "<span style='color: red;'>✘</span>"

            if Teleports.teleports_list.length > 0
                teleport_text = "<span style='color: green;'>✔</span> Place teleport points (#{Teleports.teleports_list.length})"
                teleport_tick = "<span style='color: green;'>✔</span>"
            end

            if Dollhouse.is_dollhouse_set
                dollhouse_text = "<span style='color: green;'>✔</span> Setup dollhouse view"
                dollhouse_tick = "<span style='color: green;'>✔</span>"
            end

            if Validator.poly_pass
                poly_tick = "<span style='color: green;'>✔</span>"
            end

            if Validator.surface_pass
                surface_tick = "<span style='color: green;'>✔</span>"
            end
            
            if Validator.object_pass
                object_tick = "<span style='color: green;'>✔</span>"
            end
            
            if Validator.texture_pass
                texture_tick = "<span style='color: green;'>✔</span>"
            end

            export_string = "<button class='primary-button' disabled onclick='sketchup.export_model()'>Export model</button>"
            if (Validator.is_valid)
                puts "VALIDATOR - IS VALID" if DEBUG_MODE
                export_string = "<button class='primary-button' onclick='sketchup.export_model()'>Export model</button>"
            end

            if @validation_run
                validate_string = "
                <div class = 'main-text'>#{teleport_tick} Teleport points are all in place</div>
                <div class = 'main-text'>#{dollhouse_tick} Dollhouse ready to go</div>
                <div class = 'main-text'>#{poly_tick} Number of polygons looking good #{Validator.poly_string}</div>
                <div class = 'main-text'>#{surface_tick} Total surface area is within limits #{Validator.surface_string}</div>
                <div class = 'main-text'>#{object_tick} Total number of meshes within limits #{Validator.object_string}</div>
                <div class = 'main-text'>#{texture_tick} Total number of textures within limits #{Validator.texture_string}</div>
                "
            else
                validate_string = "
                <br><br>
                <div class = 'main-text' style='text-align:center !important'>Validation has not been run since changes were made<div>
                <br><br><br><br>
                "
            end
            # puts Validator.poly_string
            # puts Validator.surface_string

            html_content = <<-HTML
                <html>
                <head>
                <style>

                    #{ENVIZSTYLES}
                    
                </style>
                </head>
                <body>
                <div class = "make-border">
                    <div id="overlay"> <button class="loading-button">Working...</button> </div>
                    <div class = "main-heading">Enviz SketchUp plugin</div>
                    <div class = "main-text">Follow the steps below to setup your model for the Enviz conversion. <a href="https://enviz-co.notion.site/Space-AI-User-Guide-945238c3971549dabc066881c13dc3e9" target="_blank">Learn more</a></div>
                    <br>
                    <div class = "sub-heading">Step 1: Clean-up your model</div>
                    <div class = "main-text">Ensure the face orientation in your model are setup correctly for export.</div>
                    <button class="primary-button" onclick="sketchup.summon_fix_faces()">Fix face orientation</button>
                    <br>
                    <div class = "sub-heading">Step 2: Setup the scene</div>
                    <div class = "main-text">Place your teleport points and setup your dollhouse view.</div>
                    <button class="primary-button" onclick="sketchup.summon_teleport_creator()">#{teleport_text}</button>
                    <button class="primary-button"  onclick="sketchup.summon_dollhouse_creator()">#{dollhouse_text}</button>
                    <br>
                    <div class = "sub-heading">Step 3: Check all is good and export</div>
                    <div class = "main-text">Check that everything is good to go with your model and export it for the client portal.</div>
                    <button class="primary-button"  onclick="sketchup.validate_model()">Validate model</button>
                    #{validate_string}
                    <hr class="colored-line">
                    #{export_string}
                    <div style = "text-align:right"><div class = "main-text">v#{Enviz_SketchUp_Plugin::VERSION}</div></div>
                </div>
                </body>
                </html>
                <script>
                    function toggleOverlay(show) {
                        document.getElementById('overlay').style.display = show ? 'flex' : 'none';
                        if (show) {
                            document.body.classList.add('disable-mouse');
                        } else {
                            document.body.classList.remove('disable-mouse');
                        }
                    }
                </script>
            HTML

            if is_update
                update_dialog(html_content, winWidth, winHeight)
            else
                summon_dialog(html_content, winWidth, winHeight)
            end

        end
        # ------------------------------------------- MAIN --------------------------------------------- #
        # ------------------------------------------- FIX FACES --------------------------------------------- #
        def self.goto_fix_faces(is_update)

            @current_menu = "fix_faces"
            

            winWidth = 450;
            winHeight = 550;

            tool_button_text1 = "<button class='primary-button' onclick='sketchup.tool_flipper()'>⚙ Reverse faces</button>";
            tool_button_text2 = "<button class='primary-button' onclick='sketchup.tool_correct_object()'>⚙ Smart orient faces</button>";

            if is_tool_active(Tool_Flipper)
                tool_button_text1 = "<button class='primary-button-tool_on' onclick='sketchup.tool_flipper()'>✓ Reversing faces</button>";
                tool_button_text2 = "<button disabled class='primary-button' onclick='sketchup.tool_correct_object()'>⚙ Smart orient faces</button>";
                result = Sketchup.status_text = "Click on a red face to flip it."
            end

            if is_tool_active(Tool_CorrectObject)
                tool_button_text1 = "<button disabled class='primary-button' onclick='sketchup.tool_flipper()'>⚙ Reverse faces</button>";
                tool_button_text2 = "<button class='primary-button-tool_on' onclick='sketchup.tool_correct_object()'>✓ Smart orient faces</button>";
                result = Sketchup.status_text = "Click on a red face to flip it."
            end

            html_content = <<-HTML
                <html>
                <head>
                <style>

                    #{ENVIZSTYLES}
                    
                </style>
                </head>
                <body>
                <div class = "make-border">
                    <div id="overlay"> <button class="loading-button">Working...</button> </div>
                    <button class="back-button" onclick="sketchup.summon_main()">< Back to Enviz SketchUp plugin</button>
                    <hr class="colored-line">
                    <div class = "main-heading">Fix face orientation</div>
                    <div class = "main-text">For optimal performance, the Enviz platform only renders the visible side of each surface.  Occasionally surfaces will be incorrectly oriented and need to be reversed or will appear transparent.<br><br>
                    Activate the tool and if any faces appear red throughout your model, use the tools to re-orient them.<br><br>
                    </div>
                    #{tool_button_text1}
                    #{tool_button_text2}
                    <hr class="colored-line">
                    <button class="final-button" onclick="sketchup.summon_main()">Done</button>
                    <div style = "text-align:right"><div class = "main-text">v#{Enviz_SketchUp_Plugin::VERSION}</div></div>
                </div>
                </body>
                </html>
            HTML

            if is_update
                update_dialog(html_content, winWidth, winHeight)
            else
                summon_dialog(html_content, winWidth, winHeight)
            end

        end
        # ------------------------------------------- FIX FACES --------------------------------------------- #
        # ------------------------------------------- DOLLHOUSE CREATOR --------------------------------------------- #
        def self.goto_dollhouse_creator(is_update)

            @current_menu = "dollhouse_creator"

            winWidth = 450;
            winHeight = 360;

            tool_button_text1 = "
            <div class = 'main-heading'>Setup dollhouse view</div>
            <div class = 'main-text'>Dollhouse View creates an interactive miniature of your space.<br><br>
            Please set the pivot point and angle for camera rotation.<br><br>
            For best results this should be central to your model and on the ground floor.</div>
            <button class='primary-button' onclick='sketchup.tool_dollhouse_creator_pivot()'>⚙ Set pivot point</button>
            ";

            if is_tool_active(Tool_Dollhouse_Creator_Pivot)
                winHeight = 380;
                tool_button_text1 = "
                <div class = 'main-heading'>Setup dollhouse view</div>
                <div class = 'sub-heading'>Set pivot point</div>
                <div class = 'main-text'>Dollhouse View creates an interactive miniature of your space.<br><br>
                Please set the pivot point and angle for camera rotation.<br><br>
                For best results this should be central to your model and on the ground floor.</div>
                <button class='primary-button-tool_on' onclick='sketchup.tool_dollhouse_creator_pivot()'>✓ Click to set pivot point</button>
                ";
            end

            if is_tool_active(Tool_Dollhouse_Creator_View)
                winHeight = 605;
                tool_button_text1 = "
                <div class = 'main-heading'>Setup dollhouse view</div>
                <div class = 'container'>
                    <div class='main-text'>✓ Pivot point set</div>
                </div>
                <hr class='colored-line'>
                <div class = 'sub-heading'>Set camera start point</div>
                <div class = 'main-text'>Set the camera start point for your dollhouse.</div>
                <div class = 'main-text'>This is the initial point where the camera will start rotating from.</div>
                <br>
                <div class='slider-container'>
                    <label class = 'side-text'>Rotation</label><br>
                    <input type='range' class='range_slider' id='rotation_input' min='0' max='360' value='180'>
                </div>
                <div class='slider-container'>
                    <label class = 'side-text'>Distance</label><br>
                    <input type='range' class='range_slider' id='distance_input' min='5' max='100' value='60'>
                </div>
                <div class='slider-container'>
                    <label class = 'side-text'>Height</label><br>
                    <input type='range' class='range_slider' id='height_input' min='3' max='100' value='50'>
                </div><br>
                <button class='primary-button' onclick='sketchup.tool_confirm_dollhouse()'>+ Confirm Dollhouse View</button>

                <script>
                    function updateDollhouseView() {
                        const rotation = document.getElementById('rotation_input').value;
                        const distance = document.getElementById('distance_input').value;
                        const height = document.getElementById('height_input').value;
                        window.sketchup.update_dollhouse_view(rotation, distance, height);
                    }

                    updateDollhouseView();
                    
                    document.addEventListener('DOMContentLoaded', function() {
                        document.getElementById('rotation_input').addEventListener('input', updateDollhouseView);
                        document.getElementById('distance_input').addEventListener('input', updateDollhouseView);
                        document.getElementById('height_input').addEventListener('input', updateDollhouseView);
                    });
                </script>
                ";
            end

            html_content = <<-HTML
                <html>
                <head>
                <style>

                    #{ENVIZSTYLES}
                    
                </style>
                </head>
                <body>
                <div class = "make-border">
                    <div id="overlay"> <button class="loading-button">Working...</button> </div>
                    <button class="back-button" onclick="sketchup.summon_main()">< Back to Enviz SketchUp plugin</button>
                    <hr class="colored-line">
                    #{tool_button_text1}
                    <div style = "text-align:right"><div class = "main-text">v#{Enviz_SketchUp_Plugin::VERSION}</div></div>
                </div>
                </body>
                </html>
            HTML

            if is_update
                update_dialog(html_content, winWidth, winHeight)
            else
                summon_dialog(html_content, winWidth, winHeight)
            end

        end
        def self.update_dollhouse_view(rotation, distance, height)
            puts "Updating Dollhouse View with Rotation: #{rotation}, Distance: #{distance}, Height: #{height}" if DEBUG_MODE
            # Add your logic here to update the view based on these values
        end
        # ------------------------------------------- DOLLHOUSE CREATOR --------------------------------------------- #
        # ------------------------------------------- TELEPORT TOOL --------------------------------------------- #
        def self.goto_teleport_creator(is_update)

            # REMOVED: 09/05/2024
            # 
            # <div class = 'sub-heading'>Select teleport point</div>
            #     <br>
            #     <div class = 'squishy-text'>Selected teleport point</div>
            #     <div class = 'container'>
            #         <select class='dropdown' id='teleport-select'>
            #             #{teleport_list}
            #         </select>
            #         <button class='primary-button' onclick=onPreview()>👁 Preview</button>
            #         <button class='trash-button' onclick='sketchup.delete_teleport()'>🗑</button>
            #     </div>
            # <hr class='colored-line'>

            @current_menu = "teleport_creator"

            winWidth = 450;
            winHeight = 410;

            tool_button_text1 = "
            <button class='primary-button' onclick='sketchup.tool_teleport_creator()'>+ Place teleport point</button>
            <button class='primary-button' onclick='sketchup.manage_teleports()'>Manage teleport points</button>
            ";
            if is_tool_active(Teleports::Tool_Teleport_Creator)
                tool_button_text1 = "
                <button class='primary-button-tool_on' onclick='sketchup.tool_teleport_creator()'>✓ Click to place teleport</button>
                <button class='primary-button' onclick='sketchup.manage_teleports()'>Manage teleport points</button>
                ";
            end

            Teleports.update_teleports_list

            teleport_list = ""
            Teleports.teleports_list.each do |teleport|
                # puts teleport.name
                teleport_list << "<option value='#{teleport.name}'>#{teleport.name}</option>\n"
            end

            selected_section = ""
            selection = Sketchup.active_model.selection
            if selection.length == 1
                entity = selection[0]
                if entity.get_attribute('EnvizAttributes', 'TeleportType') # = teleport_type == 'EnvizTeleport'
                    teleport_type = entity.get_attribute('EnvizAttributes', 'TeleportType')
                    teleport_order = entity.get_attribute('EnvizAttributes', 'TeleportOrder')
                    teleport_rotation = entity.transformation.xaxis.angle_between(X_AXIS).radians.to_i
                    puts teleport_rotation if DEBUG_MODE
                    if teleport_type == 'EnvizTeleport'
                        winHeight = 700
                        selected_section = "
                        <hr class='colored-line'>
                        <div class = 'sub-heading'>Modify teleport point</div>
                        <div class = 'main-text'>Use the tools below to modify the selected teleport point.</div>
                        <br>
                        <div class = 'squishy-text'>Name teleport point</div>
                        <div class='container'>
                            <input type='text' class='text_input' id='teleport_name_input' value='#{entity.name}' onchange='sketchup.set_teleport_name(value)'>
                            <button class='primary-button' onclick=sketchup.teleport_preview(#{teleport_order})>👁 Preview</button>
                            <button class='trash-button' onclick='sketchup.delete_teleport(#{teleport_order})'>🗑</button>
                        </div>
                        <hr class='colored-line'>
                        <div class = 'squishy-text'>Rotate teleport point</div>
                        <div class = 'container'>
                            <input type='text' class='text_input' id='rotation_input' value='#{teleport_rotation}'>
                            <input type='range' class='range_slider' id='range_input' min='0' max='360' value='#{teleport_rotation}'>
                        </div>
                        
                        <script>
                            
                            var text_input = document.getElementById('rotation_input');
                            var slider = document.getElementById('range_input');
                            var intervalId = null;

                            function startIncrementing() {
                                if (intervalId !== null) return; // Prevent multiple intervals

                                intervalId = setInterval(function() {
                                    var slider_val = Number(slider.value);
                                    text_input.value = slider_val;
                                    window.sketchup.handle_rotation_change(slider_val);
                                }, 10); // Every 0.1 seconds
                            }

                            function stopIncrementing() {
                                if (intervalId !== null) {
                                clearInterval(intervalId);
                                intervalId = null;
                                }
                            }

                            slider.addEventListener('mousedown', startIncrementing);
                            document.addEventListener('mouseup', stopIncrementing);
                            slider.addEventListener('touchstart', startIncrementing);
                            document.addEventListener('touchend', stopIncrementing);

                            function onPreview() {
                                window.sketchup.teleport_preview();
                            }

                        </script>

                        <script>
                            document.getElementById('rotation_input').addEventListener('input', function() {
                                var rotationValue = parseFloat(this.value); // Ensure the value is treated as a number

                                // Check if the value is NaN and set to 0 if true
                                if (isNaN(rotationValue)) {
                                    rotationValue = 0;
                                }

                                // Clamp the rotation value between 0 and 360 degrees
                                rotationValue = Math.max(0, Math.min(360, rotationValue));

                                // Update the value of the input field in case it was clamped or set to 0
                                this.value = rotationValue;

                                // Call the SketchUp Ruby function
                                window.sketchup.handle_rotation_change(rotationValue);

                                // Update the slider value to match the clamped rotation input
                                var slider = document.getElementById('range_input');
                                slider.value = rotationValue;
                            });
                        </script>
                        "
                    end
                end
            end

            html_content = <<-HTML
                <html>
                <head>
                <style>

                    #{ENVIZSTYLES}
                    
                </style>
                </head>
                <body>
                <div class = "make-border">
                    <div id="overlay"> <button class="loading-button">Working...</button> </div>
                    <button class="back-button" onclick="sketchup.summon_main()">< Back to Enviz SketchUp plugin</button>
                    <hr class="colored-line">
                    <div class = "main-heading">Place teleport points</div>
                    <div class = "main-text">Teleport points are required for several navigation features on Enviz.<br><br>
                    Please place a teleport point central to each of the primary spaces and name them accordingly i.e. 'kitchen', 'bedroom 1' etc.</div>
                    #{tool_button_text1}
                    #{selected_section}
                    <div style = "text-align:right"><div class = "main-text">v#{Enviz_SketchUp_Plugin::VERSION}</div></div>
                </div>
                </body>
                </html>
            HTML

            if is_update
                update_dialog(html_content, winWidth, winHeight)
            else
                summon_dialog(html_content, winWidth, winHeight)
            end

        end

        # ------------------------------------------- TELEPORT TOOL --------------------------------------------- #
        # ------------------------------------------- TELEPORT MANAGER --------------------------------------------- #
        def self.goto_teleport_manager(is_update)

            result = Sketchup.status_text = "Drag teleport points top to bottom in the order you want your users to experience them in Enviz."

            @current_menu = "teleport_manager"

            winWidth = 450;
            winHeight = 800;

            Teleports.update_teleports_list

            teleport_list = ""
            Teleports.teleports_list.each do |teleport|
                # puts teleport.name
                teleport_order = teleport.get_attribute('EnvizAttributes', 'TeleportOrder')
                teleport_list << "
                <li class='draggable' draggable='true'>
                    <span class='teleport-name'>#{teleport.name}</span>
                    <div class='buttons'>
                        <button class='primary-button' onclick=sketchup.teleport_preview_manager(#{teleport_order})>👁 Preview</button>
                        <button class='trash-button' onclick='sketchup.delete_teleport_manager(#{teleport_order})'>🗑</button>
                    </div>
                </li>
                "
            end

            html_content = <<-HTML
            <!DOCTYPE html>
                <html lang="en">
                <head>
                    <meta charset="UTF-8">
                    <meta name="viewport" content="width=device-width, initial-scale=1.0">
                    <title>Drag and Drop List</title>
                    <style>
                        #{ENVIZSTYLES}
                        .draggable {
                            padding: 4px;
                            margin: 8px 0;
                            background-color: #FFFFFF;
                            border: 1px solid #6332A4; /* Added 3px border */
                            cursor: grab;
                            color: #6332A4;
                            text-align: center;
                            font-family: "Helvetica";
                            font-size: 14px;
                            font-weight: 700;
                            line-height: 100%;
                            letter-spacing: 0.32px;
                            display: flex;
                            justify-content: space-between;
                            align-items: center;
                            gap: 4px;
                            border-radius: 5px;
                            height: 70px;
                            width: 100%;
                            box-sizing: border-box;
                            filter: blur(0.25px);
                        }
                        #draggable-list {
                            padding: 0;
                            margin: 0;
                            list-style: none;
                            width: 100%; /* Ensure the list takes full width */
                            box-sizing: border-box; /* Include padding and border in the element's total width and height */
                        }
                        .dragging {
                            opacity: 0.5;
                        }
                        .placeholder {
                            background-color: #e0e0e0;
                            border: 1px dashed #ccc;
                            margin: 8px 0;
                            height: 32px;
                        }
                        .buttons {
                            display: flex;
                            gap: 5px;
                        }
                    </style>
                </head>
                <body>
                <div class = "make-border">
                    <button class="back-button" onclick="sketchup.summon_main()">< Back to Enviz SketchUp plugin</button>
                    <div class = "main-heading">Manage teleport points</div>
                    <hr class="colored-line">
                    <div class = "sub-heading">Order teleport points</div>
                    <hr class="colored-line">
                    <div class = "main-text">Drag teleport points top to bottom in the order you want your users to experience them in Enviz.<br><br>
                </div>
                    <ul id="draggable-list">
                        #{teleport_list}
                    </ul>
                    <div style = "text-align:right"><div class = "main-text">v#{Enviz_SketchUp_Plugin::VERSION}</div></div>
                
                    <input type="hidden" id="teleport-order" value="">

                    <script>
                        const list = document.getElementById('draggable-list');
                        let draggingEl;
                        let placeholder;
                        let originalOrder = [];
                        let item_to_move = -1;

                        function getElementIndex(element) {
                            return Array.from(element.parentNode.children).indexOf(element);
                        }

                        document.addEventListener('DOMContentLoaded', () => {
                            // Set data-order attributes initially and capture originalOrder
                            Array.from(list.children).forEach((item, index) => {
                                item.dataset.order = index;
                            });
                            originalOrder = Array.from(list.children).map(item => item.dataset.order);
                        });

                        list.addEventListener('dragstart', (e) => {
                            draggingEl = e.target;
                            draggingEl.classList.add('dragging');
                            placeholder = document.createElement('li');
                            placeholder.classList.add('placeholder');
                            e.dataTransfer.effectAllowed = 'move';
                            
                            // Capture the original order when dragging starts
                            originalOrder = Array.from(list.children).map(item => item.dataset.order);
                            console.log('Drag started');

                            item_to_move = getElementIndex(draggingEl);
                        });

                        list.addEventListener('dragend', () => {
                            draggingEl.classList.remove('dragging');
                            placeholder && placeholder.remove();
                            console.log('Drag ended');
                        });

                        list.addEventListener('dragover', (e) => {
                            e.preventDefault();
                            const afterElement = getDragAfterElement(list, e.clientY);
                            if (afterElement == null) {
                                list.appendChild(placeholder);
                            } else {
                                list.insertBefore(placeholder, afterElement);
                            }
                            console.log('Drag over');
                        });

                        list.addEventListener('drop', (e) => {
                            e.preventDefault();
                            placeholder.replaceWith(draggingEl);
                            console.log('Dropped');
                            // Ensure the DOM order is updated before calling updateTeleportOrder
                            requestAnimationFrame(() => {
                                const newPosition = getElementIndex(draggingEl);
                                updateTeleportOrder(item_to_move, newPosition);
                            });
                        });

                        function getDragAfterElement(container, y) {
                            const draggableElements = [...container.querySelectorAll('.draggable:not(.dragging)')];

                            return draggableElements.reduce((closest, child) => {
                                const box = child.getBoundingClientRect();
                                const offset = y - box.top - box.height / 2;
                                if (offset < 0 && offset > closest.offset) {
                                    return { offset: offset, element: child };
                                } else {
                                    return closest;
                                }
                            }, { offset: Number.NEGATIVE_INFINITY }).element;
                        }

                        function updateTeleportOrder(item_to_move, new_position) {
                            if (item_to_move != new_position){

                                window.sketchup.reorder_teleports(item_to_move, new_position);

                            }
                        }
                    </script>
                </body>
                </html>        
                HTML

            if is_update
                update_dialog(html_content, winWidth, winHeight)
            else
                summon_dialog(html_content, winWidth, winHeight)
            end

        end
        class SimpleSelectionObserver < Sketchup::SelectionObserver

            @was_teleport_selected = false

            puts "selection observer!" if DEBUG_MODE
            # Called when the selection changes, including adding to or removing from the selection
            def onSelectionBulkChange(selection)
                if selection.empty?
                puts "All selections have been cleared." if DEBUG_MODE
                else
                    if selection.length == 1
                        entity = selection[0]
                        
                        if entity.is_a?(Sketchup::Group) || entity.is_a?(Sketchup::ComponentInstance)
                            teleport_type = entity.get_attribute('EnvizAttributes', 'TeleportType')
                            if teleport_type == 'EnvizTeleport'
                            puts "Teleport selected: #{entity.name}" if DEBUG_MODE
                            @was_teleport_selected = true
                            Enviz_UI.goto_teleport_creator(true);
                            # Implement any additional logic needed when a teleport is selected
                            else
                            puts "Non-teleport entity selected" if DEBUG_MODE
                            end
                        end
                        else
                        puts "Multiple entities or no entity selected" if DEBUG_MODE
                        @was_teleport_selected = false
                    end
                end
            end

            # Called when the selection is completely cleared
            def onSelectionCleared(selection)
                if @was_teleport_selected
                    Enviz_UI.goto_teleport_creator(true);
            
                end
            end
            def onPreCloseModel(model)
                # This is a good place to do cleanup like detaching observers
                selection = model.selection
                selection.remove_observer(@simple_observer) if @simple_observer
                puts "Model is about to be closed - Observers detached." if DEBUG_MODE
            end

        end

        # ------------------------------------------- TELEPORT MANAGER --------------------------------------------- #
    end

end
