module Enviz_SketchUp_Plugin

    DEBUG_MODE = true

    VERSION = "1.1.7"

    puts "Enviz SketchUp Plugin v#{VERSION}" if DEBUG_MODE

    # Requirements
    require 'sketchup.rb'
    Sketchup::require('enviz_sketchup_plugin/enviz_sketchup_helpers')
    Sketchup::require('enviz_sketchup_plugin/enviz_small_tools')
    Sketchup::require('enviz_sketchup_plugin/enviz_ui')
    Sketchup::require('enviz_sketchup_plugin/enviz_logging')
    Sketchup::require('enviz_sketchup_plugin/teleport_handler')
    Sketchup::require('enviz_sketchup_plugin/tool_flipper')

    # ------------- TEST --------------- #  

    # no tests right now

    # ------------- TEST --------------- #  

    # ------------- MAIN --------------- #

    # Adds a button to the toolbar
    def self.add_toolbar_button()

        ObjectSpace.each_object(UI::Toolbar) do |tb| 
            if tb.name.downcase.include?("enviz")
                puts "Enviz button already present" if DEBUG_MODE
                return
            end
        end

        toolbar = UI::Toolbar.new "Enviz"

        cmd = UI::Command.new("Enviz Toolbar") {
            start_plugin
        }
        cmd.small_icon = "icon_16.png"
        cmd.large_icon = "icon_24.png"
        cmd.tooltip = "Enviz SketchUp Plugin"
        cmd.status_bar_text = "Prepare, optimize and validate your model for upload to Space AI."
        cmd.menu_text = "Enviz SketchUp Plugin"
        toolbar = toolbar.add_item cmd
        toolbar.show
        
    end

    def self.add_menu_option()

        unless file_loaded?(__FILE__)
            # Create a new menu
            my_menu = UI.menu("Plugins").add_submenu("Enviz")
          
            # Add items to the menu
            my_menu.add_item("Open Enviz SketchUp Plugin") { Enviz_UI.goto_main(false) }
          
            file_loaded(__FILE__)
        end

    end

    Enviz_Logging.start_logging
    add_toolbar_button
    add_menu_option

    def self.start_plugin
        
        Enviz_UI.attach_selection_observer
        Enviz_UI.goto_main(false)

    end
    # ------------- MAIN --------------- #

    # ------------- REFERENCE --------------- #

    # ------------ KEY FUNCTIONALITY --------------- #
    #
    # Face flipper ✓
    #    - Click to flip tool ✓
    #    - Flip all tool ✓
    # Teleport creator ✓
    #    - Create teleport ✓
    #    - Delete teleport ✓
    #    - Lock teleports from user tampering ✓
    #    - Teleport UI panel (waiting on UI) ✓
    #    - Rotate teleport (waiting on UI) ✓
    # UI ✓
    #    - Menu handler ✓
    #    - JS/Ruby passthrough ✓
    # Dollhouse creator (tools execute in sequence like a wizard) ✓
    #    - Place pivot ✓
    #    - Place dollhouse (must orbit pivot) ✓
    # Validate model ✓
    #    - Check tris ✓
    #    - Check surface ✓
    #    - Check teleports ✓
    #    - Check dollhouse ✓
    #    - result ui + unlock export ✓
    #    - any change = invalidate model ✓
    # Export ✓
    #    - Export FBX + textures ✓
    #    - Export dollhouse info ✓
    #    - Export teleport info ✓
    #    - Zip data ✓
    #    - Change extension ✓
    #

end
