Sketchup::require('enviz_sketchup_plugin/enviz_sketchup_helpers')
Sketchup::require('enviz_sketchup_plugin/enviz_ui')
module Enviz_SketchUp_Plugin
    module Dollhouse

        @dollhouse_pivot = nil
        @dollhouse_view_point = nil
        @dollhouse_view_rot = nil

        def self.set_dollhouse_pivot(point)
            @dollhouse_pivot = Geom::Point3d.new(point.x, point.y, point.z)
            puts "Dollhouse pivot set to: #{@dollhouse_pivot}" if DEBUG_MODE
        end

        def self.get_dollhouse_view()
            @dollhouse_view_point
        end
        def self.get_dollhouse_pivot()
            @dollhouse_pivot 
        end
        
        def self.set_dollhouse_view()
            view = Sketchup.active_model.active_view
            camera = view.camera
            
            @dollhouse_view_point = Geom::Point3d.new(camera.eye.x, camera.eye.y, camera.eye.z)
            @dollhouse_view_rot = camera.direction  # Camera direction vector
        
            dp = Dollhouse.get_dollhouse_pivot
            dv = Dollhouse.get_dollhouse_view
            puts "Dollhouse pivot point set to: #{Geom::Point3d.new(dp.x, dp.y, dp.z)}" if DEBUG_MODE
            puts "Dollhouse view point set to: #{Geom::Point3d.new(dv.x, dv.y, dv.z)}" if DEBUG_MODE
        end
        
        def self.is_dollhouse_set
            if @dollhouse_pivot && @dollhouse_view_point && @dollhouse_view_rot
                true
            else
                false
            end
        end

        def self.update_dollhouse_view(rotation, distance, height)
            # puts "Updating Dollhouse View with Rotation: #{rotation}, Distance: #{distance}, Height: #{height}"
            
            point = @dollhouse_pivot

            model = Sketchup.active_model
            view = model.active_view
            camera = view.camera

            # Convert string inputs to float
            rotation = rotation.to_f
            distance = distance.to_f * 39.37 # 1m = 39.37 inches, all system measurement should be in inches 
            # distance = Enviz_Sketchup_Helpers.convert_to_meters(distance)
            height = height.to_f * 39.37 
            # height = Enviz_Sketchup_Helpers.convert_to_meters(height)

            # puts "DISTANCE = #{distance.to_m}"

            # Calculate the new camera position
            # Convert rotation to radians
            rad_rotation = rotation * Math::PI / 180.0

            # Calculate camera position
            x = point.x + distance * Math.cos(rad_rotation)
            y = point.y + distance * Math.sin(rad_rotation)
            z = point.z + height

            dp = Dollhouse.get_dollhouse_pivot
            # puts "#{dp.x}, #{dp.y}, #{dp.z}"
            new_camera_position = Geom::Point3d.new(x, y, z)

            # Calculate the new direction vector from the camera to the point
            new_direction = point.vector_to(new_camera_position)

            # The correct direction vector is from the camera to the point, not the other way around
            correct_direction = new_camera_position.vector_to(point)

            # Set up the camera
            camera.set(new_camera_position, correct_direction, Geom::Vector3d.new(0, 0, 1))

            # Update the view
            view.refresh

        end

        # Face flipper tool for correcting flipped normals
        # When activated it will change the view mode to monochrome so that backfaces are visible
        # Then the user can single left-click on faces to flip them
        class Tool_Dollhouse_Creator_Pivot

            include Enviz_Sketchup_Helpers
                
            def onLButtonDown(flags, x, y, view)
                hit_point = Enviz_Sketchup_Helpers.raycast_hit_point(x, y, view)  # Retrieve the face using the raycast_face function
            
                if hit_point  # Check if a face was successfully retrieved and is not nil
                    Dollhouse.set_dollhouse_pivot(hit_point)
                    Sketchup.active_model.select_tool(Dollhouse::Tool_Dollhouse_Creator_View.new)
                    Enviz_UI.goto_dollhouse_creator(true)
                else
                    puts "No point selected"   if DEBUG_MODE# Optional: Notify that no face was selected
                end
            end    
        
            def activate
                Sketchup::set_status_text "Click on the scene to set dollhouse pivot"
                begin
                    puts "Ready to set dollhouse pivot" if DEBUG_MODE
                rescue Exception => e

                end
            end
        
            def deactivate(view)
                puts "Tool deactivated." if DEBUG_MODE
            end
            
        end

        class Tool_Dollhouse_Creator_View

            include Enviz_Sketchup_Helpers
                
            def onLButtonDown(flags, x, y, view)
                
            end    
        
            def activate
                begin
                    puts "Ready to set dollhouse view" if DEBUG_MODE
                    view = Sketchup.active_model.active_view
                    @original_camera = view.camera
                    @original_position = @original_camera.eye
                    @original_target = @original_camera.target
                    @original_up = @original_camera.up
                    @original_fov = @original_camera.fov
                    @original_perspective = @original_camera.perspective?

                    new_camera = Sketchup::Camera.new(@original_position, @original_target, @original_up)
                    new_camera.perspective = true
                    new_camera.fov = 60
                    view.camera = new_camera
                rescue Exception => e

                end
            end
        
            def deactivate(view)
                puts "Tool deactivated." if DEBUG_MODE
                # Restore the original camera settings using individual components
                new_camera = Sketchup::Camera.new(@original_position, @original_target, @original_up)
                new_camera.perspective = @original_camera.perspective?
                new_camera.aspect_ratio = @original_camera.aspect_ratio if @original_camera.perspective?
                new_camera.fov = @original_fov
                new_camera.perspective = @original_perspective
                view.camera = new_camera
                view.invalidate
            end
            
        end

    end
end