Sketchup::require('enviz_sketchup_plugin/enviz_sketchup_helpers')
Sketchup::require('enviz_sketchup_plugin/enviz_ui')
Sketchup::require('enviz_sketchup_plugin/Hynk_orient_faces')
require 'set'

module Enviz_SketchUp_Plugin

    # Face flipper tool for correcting flipped normals
    # When activated it will change the view mode to monochrome so that backfaces are visible
    # Then the user can single left-click on faces to flip them
    class Tool_Flipper

        include Enviz_Sketchup_Helpers

            
        def onLButtonDown(flags, x, y, view)
            selected = Enviz_Sketchup_Helpers.raycast_face(x, y, view)  # Retrieve the face using the raycast_face function
        
            if selected  # Check if a face was successfully retrieved and is not nil
                
                dir = face_normal_direction(selected)

                if dir
                
                    # face is pointing toward viewport already
                
                elsif

                    selected.reverse!

                    flip_face_textures_with_uv(selected)

                end

            else
                # Optional: Notify that no face was selected (this is slow)
                # puts "No face selected"  
            end
        end    

        def flip_face_textures_with_uv(the_face)

            ff = the_face.material
            bf = the_face.back_material

            the_face.material = bf
            the_face.back_material = ff

        end
      
        def activate

            puts "Attempting to set backfaces to red" if DEBUG_MODE
            begin
                Enviz_Sketchup_Helpers.set_backfaces_red
            rescue Exception => e
                puts "Failed to set backfaces: #{e.message}" if DEBUG_MODE
            end
            puts "Tool activated. Click on the scene to perform raycast." if DEBUG_MODE


        end
      
        def deactivate(view)
            puts "Tool deactivated." if DEBUG_MODE
            Enviz_Sketchup_Helpers.reset_render_options
            view.invalidate
        end

        def onMouseMove(flags, x, y, view)
            # Check if the left mouse button is still down
            if flags & 1 == 1
                # Call onLButtonDown method again with the same parameters
                onLButtonDown(flags, x, y, view)
            end
        end
        
    end

    class Tool_CorrectObject

        include Enviz_Sketchup_Helpers

        def onLButtonDown(flags, x, y, view)

            Enviz_UI.show_processing_overlay(true, 0)

            selected = Enviz_Sketchup_Helpers.raycast_face(x, y, view)  # Retrieve the face using the raycast_face function
        
            if selected  # Check if a face was successfully retrieved and is not nil

                dir = face_normal_direction(selected)

                if dir
                    # face is pointing toward viewport already
                elsif
                    selected.reverse!
                    
                    flip_face_textures_with_uv(selected)
                end

                Enviz_Orient_Faces.orient_faces(selected)
                #selected.orient_faces
                # visited_faces = traverse_connected_faces(selected)

            else

                # puts "No face selected" 

            end

            Enviz_UI.show_processing_overlay(false, 0.5)

        end 

        def activate
            Sketchup::set_status_text "Click on the red faces to fix them"
            puts "Attempting to set backfaces to red" if DEBUG_MODE
            begin
                Enviz_Sketchup_Helpers.set_backfaces_red
            rescue Exception => e
                puts "Failed to set backfaces: #{e.message}" if DEBUG_MODE
            end
            puts "Tool activated. Click on the scene to perform raycast." if DEBUG_MODE
        end
      
        def deactivate(view)
            puts "Tool deactivated." if DEBUG_MODE
            Enviz_Sketchup_Helpers.reset_render_options
            view.invalidate
        end

        # NOTE: Everything below here is unused now, we have a better external flipping method

        def traverse_connected_faces(start_face)
            visited_faces = Set.new
            stack = [start_face]

            while stack.any?
                current_face = stack.pop
                unless visited_faces.include?(current_face)
                    visited_faces.add(current_face)

                    # Get all edges of the current face
                    edges = current_face.edges

                    # Loop through each edge
                    edges.each do |edge|
                        # Get the faces connected to this edge, excluding the current face
                        connected_faces = edge.faces - [current_face]

                        # Check each connected face
                        connected_faces.each do |connected_face|
                            unless visited_faces.include?(connected_face)
                                flip_crawl(current_face, connected_face)
                                stack.push(connected_face)
                            end
                        end
                    end
                end
            end
          
            visited_faces.to_a
        end
          
        def flip_crawl(start_face, connected_face)
            
            model = Sketchup.active_model
            
            # Get vertices from both faces
            vertices_face1 = start_face.vertices
            vertices_face2 = connected_face.vertices
        
            # Find shared vertices and their indices
            shared_indices_face1 = []
            vertices_face1.each_with_index do |vertex, index|
                if vertices_face2.include?(vertex)
                  shared_indices_face1 << index + 1  # Store 1-based index
                end
            end
          
            # Check if there are exactly two shared vertices
            if shared_indices_face1.size == 2
                f1_ind1, f1_ind2 = shared_indices_face1
            
                f2_ind1 = -1
                f2_ind2 = -1
            
                vertices_face2.each_with_index do |vertex, index|
                    if vertices_face1[f1_ind1-1] == vertex
                        f2_ind1 = index + 1
                    end
                    
                    if vertices_face1[f1_ind2-1] == vertex
                        f2_ind2 = index + 1
                    end
                end    

                dir1 = get_dir(f1_ind1, f1_ind2, vertices_face1.size)
            
                dir2 = get_dir(f2_ind1, f2_ind2, vertices_face1.size)
            
                # puts ("dir1: #{dir1}, dir2: #{dir2}")    
            
                if dir1 == dir2
                    # puts "FLIPIT"
                    connected_face.reverse!
                    flip_face_textures_with_uv(selected)
                end
            else
                # puts "Error: The selected faces do not share exactly two vertices."
                f1_ind1 = f1_ind2 = nil
            end
        
        end

        def flip_face_textures_with_uv(the_face)

            ff = the_face.material
            bf = the_face.back_material

            the_face.material = bf
            the_face.back_material = ff

        end

        def get_dir(ind1, ind2, size)

            if (ind2 == (ind1+1)) || (ind1 == size and ind2 == 1)
                return 1
            else
                return -1
            end

        end

    end
end